let elementsCache = [];

const containsReadableChar = /[\w\d]/;
const IgnoreTags =
  "select, textarea, button, label, audio, video, dialog, embed, menu, nav, noframes, noscript, aside, object, script, style, svg, footer, #footer, .no-read-aloud, #speechify-player, .speechify-popup-container";

function isElemVisible(elem) {
  return !!(
    elem.offsetWidth ||
    elem.offsetHeight ||
    elem.getClientRects?.().length ||
    (elem.nodeType === 3 && elem.length > 0)
  );
}

function removeHiddenElems(arr) {
  return arr.filter(isElemVisible);
}

function cleanElementText(elem) {
  const tagsToIgnore = [
    "IMG",
    "NOSCRIPT",
    "PATH",
    "MATH",
    "SCRIPT",
    "STYLE",
    "SUB",
    "SUP",
    "SVG",
    "AMP-TWITTER",
    "AMP-VIDEO",
    "AMP-IFRAME",
    "FIGCAPTION",
  ];

  elem
    .querySelectorAll(tagsToIgnore.join(","))
    .forEach((elem) => elem.remove());

  return elem;
}

function parseElement(element) {
  if (element.innerHTML) {
    const cleanedText = cleanElementText(
      element.cloneNode(true)
    ).textContent.trim();

    if (!cleanedText.length) return "";

    return cleanedText;
  }
  return "";
}

function extractTextFromPage({
  ignoreCache = false,
  startingNode = document.body,
} = {}) {
  if (elementsCache.length && !ignoreCache) return elementsCache;

  elementsCache = [];
  textsToRead = [];

  let textNodes = _getPageContent(startingNode);
  // textNodes = textNodes.filter((_node) => !_isOutOfViewPort(_node));
  for (const node of textNodes) {
    const toHide = removeHiddenElems(
      Array.from(node.querySelectorAll("*"))
    ).filter(_skipReading);

    const originalStyles = toHide.map((node) => node.style.display);
    for (const node of toHide) node.style.display = "none";

    // Gets the text nodes and filters out nodes that don't have any content
    // after processing
    const textNodes = _getTexts(node, "all").filter((node) =>
      containsReadableChar.test(parseElement(node))
    );

    elementsCache.push(textNodes);
    for (const [i, node] of Object.entries(toHide))
      node.style.display = originalStyles[i];
  }
  elementsCache = elementsCache.flat(50);

  return elementsCache;
}

// PRIVATE
function _getPageContent(startingNode) {
  let textBlocks = _getAllTextBlocks(40, startingNode);
  const countChars = textBlocks.reduce(
    (sum, elem) => sum + _extractTextFromNode(elem).length,
    0
  );
  if (countChars < 1000) {
    textBlocks = _getAllTextBlocks(3, startingNode);
    const texts = textBlocks.map(_extractTextFromNode);
    let head;
    let tail;
    for (let i = 3; i < texts.length && !head; i++) {
      const dist = _getGaussian(texts, 0, i);
      if (texts[i].length > dist.mean + 2 * dist.stdev) head = i;
    }
    for (let j = texts.length - 4; j >= 0 && !tail; j--) {
      const gaussian = _getGaussian(texts, j + 1, texts.length);
      if (texts[j].length > gaussian.mean + 2 * gaussian.stdev) tail = j + 1;
    }
    if (head || tail) {
      textBlocks = textBlocks.slice(head || 0, tail);
    }
  }
  const toRead = [];
  for (const [idx, block] of textBlocks.entries()) {
    toRead.push(..._findHeadingsFor(block, textBlocks[idx - 1]));
    toRead.push(block);
  }
  return toRead;
}

function _getAllTextBlocks(threshold, startingNode = document.body) {
  const textBlocks = [];

  const skipTags = `h1, h2, h3, h4, h5, h6, p, [href], ${IgnoreTags}`;
  const isTextNode = (node) =>
    node.nodeType === 3 && node.nodeValue.trim().length >= 3;
  const isParagraph = (node) =>
    node.nodeType === 1 &&
    node.matches("p") &&
    isElemVisible(node) &&
    _extractTextFromNode(node).length >= threshold;

  const hasTextNodes = (elem) =>
    _testChildNodes(elem, isTextNode) &&
    _extractTextFromNode(elem).length >= threshold;

  const hasParagraphs = (elem) => _testChildNodes(elem, isParagraph);

  const containsTextBlocks = (elem) => {
    const childElements = Array.from(
      elem.querySelectorAll(
        `:scope > *${skipTags
          .split(", ")
          .map((skipTag) => `:not(${skipTag})`)
          .join("")}`
      )
    );
    return (
      childElements.some(hasTextNodes) ||
      childElements.some(hasParagraphs) ||
      childElements.some(containsTextBlocks)
    );
  };
  const addBlock = (elem, multi) => {
    if (multi) elem.setAttribute("data-read-aloud-multi-block", true);
    textBlocks.push(elem);
  };
  const walk = (elem) => {
    if (!elem) return;
    if (elem.matches("frame, iframe") && elem.contentDocument?.body)
      walk(elem.contentDocument.body);
    else if (elem.matches("dl")) addBlock(elem);
    else if (elem.matches("ol, ul")) {
      const items = Array.from(elem.children);
      if (items.some(hasTextNodes)) addBlock(elem);
      else if (items.some(hasParagraphs)) addBlock(elem, true);
      else if (items.some(containsTextBlocks)) addBlock(elem, true);
    } else if (elem.matches("tbody")) {
      const rows = Array.from(elem.children);
      if (rows.length > 3 || (rows.length > 0 && rows[0].children.length > 3)) {
        if (rows.some(containsTextBlocks)) addBlock(elem, true);
      } else rows.forEach(walk);
    } else if (hasTextNodes(elem)) addBlock(elem);
    else if (hasParagraphs(elem)) addBlock(elem, true);
    else {
      elem
        .querySelectorAll(
          `:scope > *${skipTags
            .split(", ")
            .map((skipTag) => `:not(${skipTag})`)
            .join("")}`
        )
        .forEach(walk);
    }
  };
  walk(startingNode);
  return textBlocks.filter(
    (elem) => isElemVisible(elem) && elem.getBoundingClientRect().left >= 0
  );
}

function _getTexts(elem, op) {
  if (elem.getAttribute("data-read-aloud-multi-block")) {
    elem.removeAttribute("data-read-aloud-multi-block");
    const children = removeHiddenElems(Array.from(elem.children));
    const textNodes = [];
    for (let i = 0; i < children.length; i++) {
      if (children[i].querySelectorAll("p, ul, ol, li").length > 0) {
        textNodes.push(_getTexts(children[i], op));
        continue;
      }
      textNodes.push(children[i]);
    }
    return textNodes.filter(Boolean).flat();
  }

  if (elem.querySelectorAll("p, ul, ol, li").length > 0) {
    elem.setAttribute("data-read-aloud-multi-block", true);
    return _getTexts(elem, op);
  }

  return [elem];
}

function _skipReading(elem) {
  const { float, position } = globalThis.getComputedStyle(elem);
  return (
    elem.matches(IgnoreTags) ||
    elem.matches("sup") ||
    float === "right" ||
    position === "fixed"
  );
}

function _isOutOfViewPort(elem) {
  // Get element's bounding
  const bounding = elem.getBoundingClientRect();

  // Check if it's out of the viewport on each side
  const out = {};
  // out.top = bounding.top < 0;
  out.left = bounding.left < 0;
  // out.bottom = bounding.bottom > (window.innerHeight || document.documentElement.clientHeight);
  out.right =
    bounding.right >
    (window.innerWidth || document.documentElement.clientWidth);
  // out.any = out.top || out.left || out.bottom || out.right;
  // out.all = out.top && out.left && out.bottom && out.right;

  return out.left || out.right;
}

function _getGaussian(texts, start, end) {
  start = start ?? 0;
  end = end ?? texts.length;
  let sum = 0;
  for (let i = start; i < end; i++) sum += texts[i].length;
  const mean = sum / (end - start);
  let letiance = 0;
  for (let i = start; i < end; i++)
    letiance += (texts[i].length - mean) * (texts[i].length - mean);
  return { mean, stdev: Math.sqrt(letiance) };
}

function _findHeadingsFor(block, prevBlock) {
  const result = [];
  const firstInnerElem = removeHiddenElems(
    Array.from(block.querySelectorAll("h1, h2, h3, h4, h5, h6, p"))
  )[0];
  let currentLevel = _getHeadingLevel(firstInnerElem);
  let node = _previousNode(block, true);
  while (node && node !== prevBlock) {
    const ignore = node.nodeType === 1 && node.matches(IgnoreTags);
    if (!ignore && node.nodeType === 1 && isElemVisible(node)) {
      const level = _getHeadingLevel(node);
      if (level < currentLevel) {
        result.push(node);
        currentLevel = level;
      }
    }
    node = _previousNode(node, ignore);
  }
  return result.reverse();
}

function _previousNode(node, skipChildren) {
  if (node.nodeType === 1 && node.matches("body")) return null;
  if (node.nodeType === 1 && !skipChildren && node.lastChild)
    return node.lastChild;
  if (node.previousSibling) return node.previousSibling;
  return _previousNode(node.parentNode, true);
}

function _getHeadingLevel(elem) {
  const matches = elem && /^H(\d)$/i.exec(elem.tagName);
  return matches ? Number(matches[1]) : 100;
}

function _extractTextFromNode(elem) {
  return elem.innerText?.trim() ?? "";
}

function _testChildNodes(elem, test) {
  let child = elem.firstChild;
  while (child) {
    if (test(child)) {
      return true;
    }
    child = child.nextSibling;
  }
  return false;
}

var _extractFavIcon = function(){
    var favicon = undefined;
    var nodeList = document.getElementsByTagName("link");
    for (var i = 0; i < nodeList.length; i++)
    {
        if((nodeList[i].getAttribute("rel") == "icon")||(nodeList[i].getAttribute("rel") == "shortcut icon"))
        {
            favicon = nodeList[i].getAttribute("href");
        }
    }
    return favicon;
}

var _cleanNull = function (text) {
    return null != text ? text.replace(/^null$/g, '') : void 0;
}

var _extractPreviewImage = function(){
    var images;
    images = document.querySelectorAll("meta[property='og:image'],     meta[property='og:image:url'],     meta[itemprop=image],     meta[name='twitter:image:src'],     meta[name='twitter:image'],     meta[name='twitter:image0']");
    if (images.length > 0 && _cleanNull(images[0].getAttribute('content')))
        return _cleanNull(images[0].getAttribute('content'));
    return null;
}

function main() {
  var html = '<html>'+document.getElementsByTagName('html')[0].innerHTML+'</html>';
  var image = _extractPreviewImage();
  extractTextFromPage({
    ignoreCache: true,
    startingNode: document.body,
  });
  textsToReadJSON = JSON.stringify({
    data: elementsCache.map(parseElement),
    title: document.title,
    favicon : image == null ? _extractFavIcon() : image,
    rawHtml : html
  });
  Speechify.fetchTexts(textsToReadJSON);
}

function log(payload) {
  Speechify.log(`${payload}`);
}

main();
